import numpy as np
from astropy.nddata import block_reduce
import sys
import os
import math
from adari_core.report import AdariReportBase
from adari_core.plots.combined import CombinedPlot
from adari_core.plots.cut import CutPlot
from adari_core.plots.images import ImagePlot
from adari_core.plots.histogram import HistogramPlot
from adari_core.plots.panel import Panel
from adari_core.plots.points import LinePlot, ScatterPlot
from adari_core.plots.text import TextPlot
from adari_core.utils.utils import fetch_kw_or_default
from adari_core.utils.utils import read_idp_spectrum
from adari_core.utils.clipping import clipping_percentile
from adari_core.data_libs.master_spec_science import MasterSpecScienceReport

from . import GiraffeReportMixin


class GiraffeScienceReport(GiraffeReportMixin, MasterSpecScienceReport):
    def __init__(self):
        super().__init__("giraffe_science")
        self.stacked = False
        self.spec_lw = 1.2

    def parse_sof(self):
        sci_spec = []
        stacked_spec = None
        sky = None
        raw = None
        sci = None

        for filename, catg in self.inputs:
            if catg == "SCIENCE_RBNSPEC_IDP" and filename not in sci_spec:
                sci_spec.append(filename)
            if catg == "ESOTK_SPECTRUM_IDP_FORMAT":
                stacked_spec = filename
            if catg == "ANCILLARY.MOSSKY":
                sky = filename
            if catg == "SCIENCE":
                raw = filename
            if catg == "SCIENCE_RBNSPECTRA":
                sci = filename

        single_count = 0
        file_lists = []
        if stacked_spec is not None:
            file_lists.append(
                {
                    "science": stacked_spec,
                }
            )
            for single in sci_spec :
                file_lists.append(
                    {
                        "single%03i"%single_count: single,
                    }
                )
                single_count += 1
            self.stacked = True
        else:
            for single in sci_spec :
                file_lists.append(
                    {
                        "single%03i"%single_count: single,
                        "sky": sky,
                        "raw": raw,
                        "sci": sci,
                    }
                )
                single_count += 1
    
        return file_lists

    def generate_panels(self, **kwargs):
        panels = {}


        # Collect data for SNR vs. counts plot    
        if not self.stacked:
            snr_all = []
            mean_all = []
            for i,item in enumerate(self.hdus):
                single = self.hdus[i]["single%03i"%i]
                snr = fetch_kw_or_default(single["PRIMARY"], "ESO QC SNR", default="N/A")
                mean = fetch_kw_or_default(single["PRIMARY"], "ESO QC MEAN RED", default="N/A")
                snr_all.append(snr)
                mean_all.append(mean)
        ''' 
        if self.stacked:
            sorted_raw_name = sorted(
                    [v for v in self.hdus if any("single" in s for s in v.keys())], # Get the frames with single
                    key=lambda x: next(iter(x.values()), None)[0].header.get("ESO PRO REC1 RAW1 NAME", ""),  # Order by raw file name
            )
            first_single = list(sorted_raw_name[0].values())[0] # Input with first raw file name in alphabetical order
        '''
        # Locate the science with the lowest and highest average
        if self.stacked:
            sorted_avg_flux = sorted(
                    [v for v in self.hdus if any("single" in s for s in v.keys())], # Get the frames with single
                    key=lambda x: next(iter(x.values()), None)[0].header.get("ESO QC MEAN RED", -999.0),  # Order by average flux
            )
            lowest_flux_science = list(sorted_avg_flux[0].values())[0] # Get the science with the lowest average flux
            highest_flux_science = list(sorted_avg_flux[-1].values())[0] # Get the science with the highest average flux

        if self.stacked:
            hdus = [self.hdus[0]]
        else:
            hdus = self.hdus

        for i,item in enumerate(hdus):
            if self.stacked:
                science = hdus[i]["science"]
            else:
                science = hdus[i]["single%03i"%i]
            science_fname = os.path.basename(str(science.filename()))
            spec_wave, spec_flux, spec_flux_unit = read_idp_spectrum(science["SPECTRUM"])
            if self.stacked :
                self.snr = science["SPECTRUM"].data["SNR_REDUCED"][0]
                self.qual = science["SPECTRUM"].data["CONTRIB_REDUCED"][0]
            else:
                self.snr = science["SPECTRUM"].data["SNR"][0]
                sky = hdus[i]["sky"][0]
                skydata = sky.data.T
                n_sky = sky.data[0].size
            if self.stacked :
                p = Panel(6, 6, height_ratios=[1.5, 5, 4, 1, 1, 1.5],  y_stretch=0.6)
            else :
                p = Panel(6, 6, height_ratios=[1, 5, 1, 1, 3, 1],  y_stretch=0.7)
    
            binning = -1. # no average spectrum
    
            # Create the main spectrum plot
            low_wave = low_flux = high_wave = high_flux = None
            if self.stacked:
                binning_pix = int(len(spec_wave) / ((spec_wave[-1] - spec_wave[0]) / binning))
                low_wave, low_flux, _ = read_idp_spectrum(lowest_flux_science["SPECTRUM"])
                high_wave, high_flux, _ = read_idp_spectrum(highest_flux_science["SPECTRUM"])
     
                low_wave_binned = block_reduce(low_wave, binning_pix, func=np.mean)
                high_wave_binned = block_reduce(high_wave, binning_pix, func=np.mean)
    
            specplot, lowhighplot = self.generate_spec_plot(spec_wave, spec_flux, spec_flux_unit, 
                                               binning, low_wave, low_flux, high_wave, high_flux)

            if self.stacked:
                p.assign_plot(specplot, 0, 1, xext=6)
                p.assign_plot(lowhighplot, 0, 2, xext=6)
            else:
                skyplot = LinePlot(title="",
                                    legend = False,
                                    )
                skyplot.add_data(d=[spec_wave, skydata[n_sky-1]],
                                  color="lightgreen",
                                  linewidth = self.spec_lw,
                                  label="Sky",)
                combined = CombinedPlot(title="",
                    legend = True,
                )
                combined.add_data(skyplot, zorder=0)
                combined.add_data(specplot, zorder=1)
    
                p.assign_plot(combined, 0, 1, xext=6)
    
            # Generate S/N plot
            snrplot = self.generate_snr_plot(spec_wave, self.snr)
            p.assign_plot(snrplot, 0, 2 + self.stacked, xext=6)


            # Generate quality plot
            if self.stacked:
                qualplot = self.generate_qual_plot(spec_wave, self.qual)
                p.assign_plot(qualplot, 0, 3 + self.stacked, xext=6)
            
            # Generate sky spectrum
            else:
                skyplot = LinePlot(title="",
                                   legend = False,
                                   linewidth = self.spec_lw,
                                   )
                for i_sky in range(n_sky):
                    skyplot.add_data(d=[spec_wave, skydata[i_sky]],
                                     color = "lightgreen",
                                  label="sky"+str(i_sky),)
                skyplot.x_label = f"Wavelength (nm)"
                skyplot.y_label = "Sky"
                skyplot.set_xlim(spec_wave[0], spec_wave[-1])
                skyplot.set_ylim(clipping_percentile(skydata, 98)[0], clipping_percentile(skydata, 98)[1] * 2.5)
                p.assign_plot(skyplot, 0, 3 + self.stacked, xext=6)
        
    
            # Add small plots
            
            # Cut in x direction from raw
            if not self.stacked:
                rawdata = hdus[i]["raw"][0].data
                fullfield = ImagePlot(
                            title="Raw",
                        )
                fullfield.add_data(rawdata)
                cutX = CutPlot("y", title="Central column", x_label="x", y_label="ADU",)
                cutX.add_data(
                            fullfield,
                            fullfield.get_data_coord(fullfield.data.shape[0] // 2, "y"),
                            label="Raw",
                            color="red",
                        )
                p.assign_plot(cutX, 0, 4 + self.stacked, xext=2)
        
                # SNR vs. counts
                snr = fetch_kw_or_default(science["PRIMARY"], "ESO QC SNR", default="N/A")
                mean = fetch_kw_or_default(science["PRIMARY"], "ESO QC MEAN RED", default="N/A")
                scatter_snr = ScatterPlot(
                    title="SNR vs. counts",
                    x_label="MEAN RED",
                    y_label="SNR",
                    legend=False,
                )
                scatter_snr.add_data(
                    (mean_all,snr_all),
                    label="snr_all",
                    color="red",
                    markersize=1.,
                )
                scatter_snr.add_data(
                    ([mean],[snr]),
                    label="snr",
                    color="black",
                    markersize=5.,
                )
                scatter_snr.x_min = 0
                scatter_snr.y_min = 0
                p.assign_plot(scatter_snr, 2, 4 + self.stacked, xext=2)
        
                # Histogram SCIENCE_RBNSPECTRA        
                scidata = hdus[i]["sci"][0].data[i]
                hist = HistogramPlot(
                            master_data=scidata,
                            title="Product science RBNSPECTRA",
                            master_label="",
                            v_min=-50,
                            v_max=200,
                            legend=False,
                        )
                p.assign_plot(hist, 4, 4 + self.stacked, xext=1)
        
                # Histogram raw
                hist = HistogramPlot(
                        master_data=rawdata,
                        title="Saturation limit",
                        master_label="raw",
                        v_min=55000,
                        v_max=70000,
                    )
                p.assign_plot(hist, 5, 4 + self.stacked, xext=1)
    
    
            # Upper Text Plot
            vspace = 0.4
            t0 = TextPlot(columns=1, v_space=vspace)
            col0 = (
                    str(fetch_kw_or_default(science["PRIMARY"], "INSTRUME", default="N/A"))+" science product preview",
                    "Product: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO PRO CATG", default="N/A")),
                    "Raw file: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO PRO REC1 RAW1 NAME", default="N/A")),
                    "MJD-OBS: " + str(fetch_kw_or_default(science["PRIMARY"], "MJD-OBS", default="N/A")),
            )
            t0.add_data(col0, fontsize=13)
            p.assign_plot(t0, 0, 0, xext=1)
    
            t1 = TextPlot(columns=1, v_space=vspace)
            col1 = (
                    "Target: " + str(fetch_kw_or_default(science["PRIMARY"], "OBJECT", default="N/A")),
                    "OB ID: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO OBS ID", default="N/A")),
                    "OB NAME: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO OBS NAME", default="N/A")),
                    "TPL ID: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO TPL ID", default="N/A")),
                    "RUN ID: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO OBS PROG ID", default="N/A")),
            )
            t1.add_data(col1, fontsize=13)
            p.assign_plot(t1, 2, 0, xext=1)
    
            if self.stacked:
                fibre = str(fetch_kw_or_default(science["PRIMARY"], "STACK FPS1", default="N/A"))
            else:
                fibre = str(fetch_kw_or_default(science["PRIMARY"], "FPS", default="N/A"))
            t2 = TextPlot(columns=1, v_space=vspace)
            col2 = (
                    "Fibre number (FPS): " + fibre,
                    "Slit name: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO INS SLIT NAME", default="N/A")),
                    "Exposure mode: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO INS EXP MODE", default="N/A")),
                    "Detector speed: " + str(fetch_kw_or_default(science["PRIMARY"], "ESO DET READ SPEED", default="N/A")),
            )
            t2.add_data(col2, fontsize=13)
            p.assign_plot(t2, 4, 0, xext=1)
    
            # Bottom Text Plot
            vspace = 0.4
            t4 = TextPlot(columns=1, v_space=vspace)
            col4 = ()
            if self.stacked:
                col4 += ("Exp. time [s]: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "TEXPTIME", default="N/A"),
                         "N exposures: " + "%i (%0.1f)" % (fetch_kw_or_default(science["PRIMARY"], "NCOMBINE", default="N/A"),
                            math.sqrt(fetch_kw_or_default(science["PRIMARY"], "NCOMBINE", default="N/A"))),
                         "Avg. contribution: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC CONTRIB AVG", default="N/A"),
                         "S/N average: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC SNR AVG", default="N/A"),
                )
            else:
                col4 += ("Exp. time [s]: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "TEXPTIME", default="N/A"),
                         "N exposures: " + "%i" % fetch_kw_or_default(science["PRIMARY"], "NCOMBINE", default="N/A"),
                         "S/N average: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC SNR", default="N/A"),
                         "Flux average: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC MEAN RED", default="N/A"),
                )
            t4.add_data(col4, fontsize=13)
            p.assign_plot(t4, 0, 5, xext=1)
    
            t5 = TextPlot(columns=1, v_space=vspace)
            col5 = ()
            if self.stacked:
                col5 += (
                        "Avg. S/N all single: " + "%.1f" % (fetch_kw_or_default(science["PRIMARY"], "ESO QC SNR AVG", default=0) /
                             fetch_kw_or_default(science["PRIMARY"], "ESO QC SNR RATIO", 1)),
                        "Avg. S/N improvement: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC SNR RATIO", "N/A"),
                        "Lambda start [nm]: " + "%.5f" % fetch_kw_or_default(science["PRIMARY"], "WAVELMIN", default="N/A"),
                        "Lambda end [nm]: " + "%.5f" % fetch_kw_or_default(science["PRIMARY"], "WAVELMAX", default="N/A"),
                )
            else:
                col5 += (
                        "Mag (user defined): " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC MAG", default="N/A"),
                        "Lambda bin [nm]: " + "%.5f" % fetch_kw_or_default(science["PRIMARY"], "SPEC_BIN", default="N/A"),
                        "Spectral resolution: " + "%i" % fetch_kw_or_default(science["PRIMARY"], "SPEC_RES", default="N/A"),
                        "Seeing: " + "%.2f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC FWHM", default="N/A"),
                )
            t5.add_data(col5, fontsize=13)
            p.assign_plot(t5, 2, 5, xext=1)
    
            t6 = TextPlot(columns=1, v_space=vspace)
            if self.stacked:
                col6 = (
                        "Lambda bin [nm]: " + "%.5f" % fetch_kw_or_default(science["PRIMARY"], "SPEC_BIN", default="N/A"),
                        "Spectral resolution: " + "%.5f" % fetch_kw_or_default(science["PRIMARY"], "SPEC_RES", default="N/A"),
                        "Seeing: " + "%.2f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC IA FWHM AVG", default="N/A"),
                )
            else:
                col6 = (
                        "Airmass: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC AIRM", default="N/A"),
                        "Delta t calib. [d]: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC DELTA TIME", default="N/A"),
                        "Delta T calib [C]: " + "%.1f" % fetch_kw_or_default(science["PRIMARY"], "ESO QC DELTA TEMP", default="N/A"),
                        "Score bits: " + str(fetch_kw_or_default(science["PRIMARY"], "QC_FLAG", default="N/A")),
                )
            t6.add_data(col6, fontsize=13)
            p.assign_plot(t6, 4, 5, xext=1)


            if self.stacked:
                input_files = [science.filename(),]
            else:
                input_files = [
                           science.filename(), 
                           hdus[i]["sci"].filename(),
                           hdus[i]["sky"].filename(), 
                           hdus[i]["raw"].filename(),
                ]
            addme = {
                "report_name": f"GIRAFFE_{str(science_fname).removesuffix('.fits').lower()}",
                "report_description": f"Science panel",
                "report_tags": [],
                "report_prodcatg" : "ANCILLARY.PREVIEW",
                "input_files": input_files,
            }
    
            panels[p] = addme
    
        return panels


rep = GiraffeScienceReport()

